# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import str
from builtins import object
import numpy as np
import logging
from collections import defaultdict

logger = logging.getLogger("nomad")
metainfo_section_prefix = "x_cp2k_section_input_"
metainfo_data_prefix = "x_cp2k_input_"


class CP2KInput(object):
    """The contents of a CP2K simulation including default values and default
    units from the version-specific xml file.
    """

    def __init__(self, root_section):
        self.root_section = root_section

    @staticmethod
    def decode_cp2k_unit(unit):
        """Given a CP2K unit name, decode it as Pint unit definition.
        """
        map = {
            # Length
            "bohr": "bohr",
            "m": "meter",
            "pm": "picometer",
            "nm": "nanometer",
            "angstrom": "angstrom",

            # Angle
            "rad": "radian",
            "deg": "degree",

            #Energy
            "Ry": "rydberg"
        }
        pint_unit = map.get(unit)
        if pint_unit:
            return pint_unit
        else:
            logger.error("Unknown CP2K unit definition '{}'.".format(unit))

    def set_parameter(self, path, value):
        parameter, section = self.get_parameter_and_section(path)
        if section is None:
            message = "The CP2K input does not contain a section {}".format(path)
            logger.warning(message)
        if parameter is None:
            message = "The CP2K input section {} does not contain a SECTION_PARAMETER".format(path)
            logger.warning(message)
        else:
            parameter.value = value

    def set_keyword(self, path, value, full):
        keyword, section = self.get_keyword_and_section(path)
        # If keyword found, put data in there
        if keyword and section:
            keyword.value = value
        # Keyword not found in the input tree, assuming it is a default keyword
        elif section is not None:
            split_path = path.rsplit("/", 1)
            keyword = split_path[1]
            if section.default_keyword is not None:
                section.default_keyword.value += full + "\n"
            else:
                message = "The CP2K input does not contain the keyword {}, and there is no default keyword for the section {}".format(path, split_path[0])
                logger.warning(message)

    def get_section(self, path):
        split_path = path.split("/")
        section = self.root_section
        for part in split_path:
            section = section.get_subsection(part)
            if not section:
                message = "The CP2K input does not contain the section {}".format(path)
                logger.warning(message)
                return None
        return section

    def get_section_list(self, path):
        split_path = path.split("/")
        last_section = split_path[-1]
        split_path.pop()
        section = self.root_section
        for part in split_path:
            section = section.get_subsection(part)
            if not section:
                message = "The CP2K input does not contain the section {}".format(path)
                logger.warning(message)
                return None
        return section.get_subsection_list(last_section)

    def get_keyword_and_section(self, path):
        split_path = path.rsplit("/", 1)
        keyword = split_path[1]
        section_path = split_path[0]
        section = self.get_section(section_path)

        if section is None:
            message = "The CP2K input does not contain the section {}".format(path)
            logger.warning(message)
            return (None, None)

        keyword = section.get_keyword_object(keyword)
        if keyword and section:
            return (keyword, section)
        else:
            return (None, section)

    def get_keyword(self, path, raw=False, allow_default=True):

        keyword, section = self.get_keyword_and_section(path)
        if keyword:
            return keyword.get_value(raw, allow_default)

    def set_section_accessed(self, path):
        section = self.get_section(path)
        if section:
            section.accessed = True
        else:
            message = "The CP2K input does not contain the section {}".format(path)
            logger.warning(message)

    def get_default_unit(self, path):
        keyword, section = self.get_keyword_and_section(path)
        if keyword:
            return keyword.default_unit

    def get_unit(self, path):
        keyword, section = self.get_keyword_and_section(path)
        if keyword:
            return keyword.get_unit()

    def get_parameter_and_section(self, path):
        section = self.get_section(path)
        if section is None:
            return (None, None)
        if section.section_parameter is not None:
            parameter = section.section_parameter
            parameter.accessed = section.accessed
            return (parameter, section)
        else:
            return (None, section)

    def get_parameter(self, path):
        parameter, section = self.get_parameter_and_section(path)
        if parameter:
            if parameter.value:
                return parameter.value
            elif section and section.accessed:
                return parameter.lone_value


class Section(object):
    """An input section in a CP2K calculation.
    """
    __slots__ = ['accessed', 'name', 'keywords', 'default_keyword_names', 'default_keyword', 'section_parameter', 'sections', 'description']

    def __init__(self, name):
        self.accessed = False
        self.name = name
        self.keywords = defaultdict(list)
        self.default_keyword_names = []
        self.default_keyword = None
        self.section_parameter = None
        self.sections = defaultdict(list)
        self.description = None

    def get_keyword_object(self, name):
        keyword = self.keywords.get(name)
        if keyword:
            if len(keyword) == 1:
                return keyword[0]
            else:
                logger.error("The keyword '{}' in '{}' does not exist or has too many entries.".format(name, self.name))

    def get_keyword(self, name, raw=False, allow_default=True):
        """Returns the keyword value for the given name.

        Args:
            name: The name of the keyword
            raw: Boolean indicating if the raw value (not modified in any way)
                should be returned.
            allow_default: Boolean indicating if it is allowed to return the
                default value is no actual value was set by the user in the input.
        """
        keyword_object = self.get_keyword_object(name)
        return keyword_object.get_value(raw, allow_default)

    def get_subsection(self, name):
        subsection = self.sections.get(name)
        if subsection:
            if len(subsection) == 1:
                return subsection[0]
            else:
                logger.error("The subsection '{}' in '{}' has too many entries.".format(name, self.name))
        else:
            logger.error("The subsection '{}' in '{}' does not exist.".format(name, self.name))

    def get_subsection_list(self, name):
        subsection = self.sections.get(name)
        return subsection

    def get_section_parameter(self):
        """Get the section parameter, or if not specified the lone keyword
        value.
        """
        if self.section_parameter is not None:
            value = self.section_parameter.value
            if value is None:
                value = self.section_parameter.lone_keyword_value
            return value.upper()


class InputObject(object):
    """Base class for all kind of data elements in the CP2K input.
    """
    __slots__ = ['name', 'value', 'default_value', 'description', 'data_type', 'data_dimension']

    def __init__(self, name):
        self.name = name
        self.value = None
        self.description = None
        self.data_type = None
        self.data_dimension = None
        self.default_value = None


class Keyword(InputObject):
    """Information about a keyword in a CP2K calculation.
    """
    __slots__ = ['unit', 'value_no_unit', 'default_unit', 'default_name']

    def __init__(self, name, default_value,  default_unit, default_name):
        super(Keyword, self).__init__(name)
        self.unit = None
        self.value_no_unit = None
        self.default_unit = default_unit
        self.default_value = default_value
        self.default_name = default_name

    def get_value(self, raw=False, allow_default=True):
        if raw:
            return self._get_value_raw()
        else:
            return self._get_value_formatted(allow_default)

    def _get_value_raw(self):
        """Returns the unformatted value of this keyword. This is exactly what
        was set by the used in the input as a string.
        """
        return self.value

    def _get_value_formatted(self, allow_default=False):
        """Returns the value stored in this keyword by removing the possible
        unit definition and formatting the string into the correct data type.
        If asked, will use the default value if not actual value was set by
        user.
        """
        # Decode the unit and the value if not done before
        proper_value = None
        if self.default_unit is not None:
            if self.value_no_unit is None:
                self.decode_cp2k_unit_and_value()
        if self.value_no_unit is not None:
            proper_value = self.value_no_unit
        else:
            proper_value = self.value

        # if allow_default:
        if proper_value is None:
            proper_value = self.default_value
        if proper_value is None:
            return None

        returned = None
        dim = int(self.data_dimension)
        splitted = proper_value.split()
        if len(splitted) != dim:
            logger.error("The dimensions of the CP2K input parameter {} do not match the specification in the XML file.".format(self.name))

        if dim == 1:
            try:
                if self.data_type == "integer":
                    returned = int(proper_value)
                elif self.data_type == "real":
                    returned = float(proper_value)
                elif self.data_type == "word":
                    returned = str(proper_value)
                elif self.data_type == "keyword":
                    returned = str(proper_value)
                elif self.data_type == "string":
                    returned = str(proper_value)
                elif self.data_type == "logical":
                    returned = str(proper_value)
                else:
                    logger.error("Unknown data type '{}'".format(self.data_type))
                    return
            except TypeError:
                logger.error("The CP2K input parameter {} could not be converted to the type specified in the XML file.".format(self.name))
                return
        else:
            try:
                if self.data_type == "integer":
                    returned = np.array([int(x) for x in splitted])
                elif self.data_type == "real":
                    returned = np.array([float(x) for x in splitted])
                elif self.data_type == "word":
                    returned = np.array([str(x) for x in splitted])
                elif self.data_type == "keyword":
                    returned = np.array([str(x) for x in splitted])
                elif self.data_type == "string":
                    returned = np.array([str(x) for x in splitted])
                elif self.data_type == "logical":
                    returned = np.array([str(x) for x in splitted])
                else:
                    logger.error("Unknown data type '{}'".format(self.data_type))
                    return
            except TypeError:
                logger.error("The CP2K input parameter {} could not be converted to the type specified in the XML file.".format(self.name))
                return

        return returned

    def determine_value_and_unit(self):
        """If the units of this value can be changed, return a value and the
        unit separately.
        """
        if self.default_unit is not None:
            if self.value_no_unit is None:
                self.decode_cp2k_unit_and_value()
            return self.value_no_unit
        else:
            return self.value

    def get_unit(self):

        if self.unit is None:
            self.decode_cp2k_unit_and_value()
            if self.unit is not None:
                return self.unit
            elif self.default_unit is not None:
                return self.default_unit
        else:
            return self.unit

        logger.error("The keyword '{}' does not have a unit.".format(self.default_name))
        return None

    def decode_cp2k_unit_and_value(self):
        """Given a CP2K unit name, decode it as Pint unit definition.
        """
        if self.value is not None:
            splitted = self.value.split(None, 1)
            unit_definition = splitted[0]
            if unit_definition.startswith('[') and unit_definition.endswith(']'):
                unit_definition = unit_definition[1:-1]
                self.unit = CP2KInput.decode_cp2k_unit(unit_definition)
                self.value_no_unit = splitted[1]
            elif self.default_unit:
                logger.debug("No special unit definition found, returning default unit.")
                self.unit = CP2KInput.decode_cp2k_unit(self.default_unit)
                self.value_no_unit = self.value
            else:
                logger.debug("The value has no unit, returning bare value.")
                self.value_no_unit = self.value


class SectionParameters(InputObject):
    """Section parameters in a CP2K calculation.

    Section parameters are the short values that can be added right after a
    section name, e.g. &PRINT ON, where ON is the section parameter.
    """
    __slots__ = ['lone_keyword_value', 'accessed']

    def __init__(self, default_value, lone_keyword_value):
        super(SectionParameters, self).__init__("SECTION_PARAMETERS")
        self.default_value = default_value
        self.lone_keyword_value = lone_keyword_value
        self.accessed = None

    def get_value(self):
        """Returns the value for this section parameter. Uses the user given
        value primarily, if the section is not used, return the default value
        and if the section is defined but without explicit section parameter
        returns the lone keyword value.
        """
        if self.value is not None:
            return self.value
        elif self.accessed:
            return self.lone_keyword_value
        else:
            return self.default_value


class DefaultKeyword(InputObject):
    """Default keyword in the CP2K input.
    """
    __slots__ = ['lone_value']

    def __init__(self):
        super(DefaultKeyword, self).__init__("DEFAULT_KEYWORD")
        self.lone_value = None
        self.value = ""



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
import logging
import datetime
import calendar
import numpy as np
from nomadcore.baseclasses import CommonParser
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.unit_conversion.unit_conversion import convert_unit
from .inputparser import CPMDInputParser
logger = logging.getLogger("nomad")


class CPMDCommonParser(CommonParser):
    """
    This class is used to store and instantiate common parts of the
    hierarchical SimpleMatcher structure used in the parsing of a CPMD
    calculation.
    """
    def __init__(self, parser_context):
        super(CPMDCommonParser, self).__init__(parser_context)

        #=======================================================================
        # Globally cached values
        self.cache_service.add("initial_positions", single=False, update=False)
        self.cache_service.add("atom_labels", single=False, update=False)
        self.cache_service.add("number_of_atoms", single=False, update=False)
        self.cache_service.add("simulation_cell", single=False, update=False)
        self.cache_service.add("n_steps")
        self.cache_service.add("ensemble_type")
        self.cache_service.add("time_step_ions")

        self.cache_service.add("trajectory_range", False)
        self.cache_service.add("trajectory_sample", False)
        self.cache_service.add("print_freq", 1)
        self.cache_service.add("configuration_periodic_dimensions", single=False, update=False)

        self.cache_service.add("single_configuration_calculation_to_system_ref", single=False, update=True)
        self.cache_service.add("single_configuration_to_calculation_method_ref", single=False, update=False)

    #===========================================================================
    # Common SimpleMatchers
    def header(self):
        """Returns the simplematcher that parser the CPMD header containng the
        starting information. Common to all run types.
        """
        return SM( " PROGRAM CPMD STARTED AT",
            forwardMatch=True,
            sections=["x_cpmd_section_start_information"],
            subMatchers=[
                SM( " PROGRAM CPMD STARTED AT: (?P<x_cpmd_start_datetime>{})".format(self.regexs.eol)),
                SM( "\s+VERSION (?P<program_version>\d+\.\d+)"),
                SM( r"\s+\*\*\*  (?P<x_cpmd_compilation_date>[\s\w\-:\d]+)  \*\*\*$"),
                SM( " THE INPUT FILE IS:\s+(?P<x_cpmd_input_filename>{})".format(self.regexs.eol)),
                SM( " THIS JOB RUNS ON:\s+(?P<x_cpmd_run_host_name>{})".format(self.regexs.eol)),
                SM( " THE PROCESS ID IS:\s+(?P<x_cpmd_process_id>{})".format(self.regexs.int)),
                SM( " THE JOB WAS SUBMITTED BY:\s+(?P<x_cpmd_run_user_name>{})".format(self.regexs.eol)),
            ]
        )

    def method(self):
        """Returns the simplematcher that parses informatio about the method
        used. Common to all run types.
        """
        return SM( "(?: SINGLE POINT DENSITY OPTIMIZATION)|(?: OPTIMIZATION OF IONIC POSITIONS)|(?: CAR-PARRINELLO MOLECULAR DYNAMICS)|(?: BORN-OPPENHEIMER MOLECULAR DYNAMICS)",
            subMatchers=[
                SM( " USING SEED",
                    forwardMatch=True,
                    sections=["x_cpmd_section_run_type_information"],
                    subMatchers=[
                        SM( " USING SEED\s+{}\s+TO INIT. PSEUDO RANDOM NUMBER GEN.".format(self.regexs.int)),
                        SM( " PATH TO THE RESTART FILES:\s+{}".format(self.regexs.eol)),
                        SM( " GRAM-SCHMIDT ORTHOGONALIZATION"),
                        SM( " ITERATIVE ORTHOGONALIZATION",
                            subMatchers={
                                SM("    MAXIT:\s+{}".format(self.regexs.int)),
                                SM("    EPS:\s+{}".format(self.regexs.float)),
                            }
                        ),
                        SM( " MAXIMUM NUMBER OF STEPS:\s+(?P<x_cpmd_max_steps>{}) STEPS".format(self.regexs.int)),
                        SM( " MAXIMUM NUMBER OF ITERATIONS FOR SC:\s+(?P<scf_max_iteration>{}) STEPS".format(self.regexs.int)),
                        SM( " PRINT INTERMEDIATE RESULTS EVERY\s+{} STEPS".format(self.regexs.int)),
                        SM( " STORE INTERMEDIATE RESULTS EVERY\s+{} STEPS".format(self.regexs.int)),
                        SM( " STORE INTERMEDIATE RESULTS EVERY\s+{} SELF-CONSISTENT STEPS".format(self.regexs.int)),
                        SM( " NUMBER OF DISTINCT RESTART FILES:\s+{}".format(self.regexs.int)),
                        SM( " TEMPERATURE IS CALCULATED ASSUMING EXTENDED BULK BEHAVIOR"),
                        SM( " FICTITIOUS ELECTRON MASS:\s+{}".format(self.regexs.float)),
                        SM( " TIME STEP FOR ELECTRONS:\s+(?P<x_cpmd_time_step_electrons__hbar_hartree_1>{})".format(self.regexs.float)),
                        SM( " TIME STEP FOR IONS:\s+(?P<x_cpmd_time_step_ions__hbar_hartree_1>{})".format(self.regexs.float)),

                        SM( " TRAJECTORIES ARE SAVED ON FILE"),
                        SM( " TRAJEC\.xyz IS SAVED ON FILE"),
                        SM( " ELECTRON DYNAMICS:"),
                        SM( " ION DYNAMICS:(?P<x_cpmd_ion_temperature_control>{})".format(self.regexs.eol)),

                        SM( " CONVERGENCE CRITERIA FOR WAVEFUNCTION OPTIMIZATION:\s+(?P<scf_threshold_energy_change__hartree>{})".format(self.regexs.float)),
                        SM( " WAVEFUNCTION OPTIMIZATION BY PRECONDITIONED DIIS"),
                        SM( " THRESHOLD FOR THE WF-HESSIAN IS\s+{}".format(self.regexs.float)),
                        SM( " MAXIMUM NUMBER OF VECTORS RETAINED FOR DIIS:\s+{}".format(self.regexs.int)),
                        SM( " STEPS UNTIL DIIS RESET ON POOR PROGRESS:\s+{}".format(self.regexs.int)),
                        SM( " FULL ELECTRONIC GRADIENT IS USED".format(self.regexs.int)),

                        SM( " CONVERGENCE CRITERIA FOR GEOMETRY OPTIMIZATION:\s+(?P<geometry_optimization_threshold_force__hartree_bohr_1>{})".format(self.regexs.float)),
                        SM( " GEOMETRY OPTIMIZATION BY (?P<x_cpmd_geo_opt_method>(?:GDIIS/BFGS)|(?:LOW-MEMORY BFGS)|(?:CONJUGATE GRADIENT)|(?:STEEPEST DESCENT))"),
                        SM( "   SIZE OF GDIIS MATRIX:\s+{}".format(self.regexs.int)),
                        SM( "GEOMETRY OPTIMIZATION IS SAVED ON FILE {}".format(self.regexs.eol)),
                        SM( " EMPIRICAL INITIAL HESSIAN (DISCO PARAMETRISATION)"),

                        SM( " SPLINE INTERPOLATION IN G-SPACE FOR PSEUDOPOTENTIAL FUNCTIONS",
                            subMatchers=[
                                SM( "    NUMBER OF SPLINE POINTS:\s+{}".format(self.regexs.int)),
                            ]
                        ),
                    ]
                ),
                SM( " EXCHANGE CORRELATION FUNCTIONALS",
                    sections=["x_cpmd_section_xc_information"],
                    subMatchers=[
                        # SM( " PROGRAM CPMD STARTED AT: (?P<x_cpmd_start_datetime>{})".format(self.regexs.eol)),
                    ]
                ),
            ]
        )

    def atoms(self):
        """Returns the simplematcher that parses system information including
        the atom labels, positions, and pseudopotentials. Common to all run
        types.
        """
        return SM( re.escape(" ***************************** ATOMS ****************************"),
            forwardMatch=True,
            subMatchers=[
                SM( re.escape(" ***************************** ATOMS ****************************"),
                    sections=["x_cpmd_section_system_information"],
                    subMatchers=[
                        SM( "   NR   TYPE        X(BOHR)        Y(BOHR)        Z(BOHR)     MBL".replace("(", "\(").replace(")", "\)"),
                            adHoc=self.ad_hoc_atom_information()
                        ),
                        SM( " CHARGE:\s+(?P<total_charge>{})".format(self.regexs.int)),
                    ]
                ),
                SM( re.escape("    |    Pseudopotential Report"),
                    sections=["x_cpmd_section_pseudopotential_information"],
                ),
                SM( re.escape(" *   ATOM       MASS   RAGGIO NLCC              PSEUDOPOTENTIAL *"),
                    sections=["x_cpmd_section_atom_kinds"],
                    subMatchers=[
                        SM( " \*\s+(?P<x_cpmd_atom_kind_label>{0})\s+(?P<x_cpmd_atom_kind_mass>{1})\s+(?P<x_cpmd_atom_kind_raggio>{1})\s+(?P<x_cpmd_atom_kind_nlcc>{0})\s+(?P<x_cpmd_atom_kind_pseudopotential_l>{0})\s+(?P<x_cpmd_atom_kind_pseudopotential_type>{0})\s+\*".format(self.regexs.word, self.regexs.float),
                            sections=["x_cpmd_section_atom_kind"],
                            repeats=True,
                        ),
                    ]
                ),
            ]
        )

    def cell(self):
        """Returns the simplematcher that parser the cell information. Common to all run types.
        """
        return SM( re.escape(" ************************** SUPERCELL ***************************"),
            sections=["x_cpmd_section_supercell"],
            subMatchers=[
                SM( " SYMMETRY:\s+(?P<x_cpmd_cell_symmetry>{})".format(self.regexs.eol)),
                SM( " LATTICE CONSTANT\(a\.u\.\):\s+(?P<x_cpmd_cell_lattice_constant>{})".format(self.regexs.float)),
                SM( " CELL DIMENSION:\s+(?P<x_cpmd_cell_dimension>{})".format(self.regexs.eol)),
                SM( " VOLUME\(OMEGA IN BOHR\^3\):\s+(?P<x_cpmd_cell_volume>{})".format(self.regexs.float)),
                SM( " LATTICE VECTOR A1\(BOHR\):\s+(?P<x_cpmd_lattice_vector_A1>{})".format(self.regexs.eol)),
                SM( " LATTICE VECTOR A2\(BOHR\):\s+(?P<x_cpmd_lattice_vector_A2>{})".format(self.regexs.eol)),
                SM( " LATTICE VECTOR A3\(BOHR\):\s+(?P<x_cpmd_lattice_vector_A3>{})".format(self.regexs.eol)),
                SM( " RECIP\. LAT\. VEC\. B1\(2Pi/BOHR\):\s+(?P<x_cpmd_reciprocal_lattice_vector_B1>{})".format(self.regexs.eol)),
                SM( " RECIP\. LAT\. VEC\. B2\(2Pi/BOHR\):\s+(?P<x_cpmd_reciprocal_lattice_vector_B2>{})".format(self.regexs.eol)),
                SM( " RECIP\. LAT\. VEC\. B3\(2Pi/BOHR\):\s+(?P<x_cpmd_reciprocal_lattice_vector_B3>{})".format(self.regexs.eol)),
                SM( " RECIP\. LAT\. VEC\. B3\(2Pi/BOHR\):\s+(?P<x_cpmd_reciprocal_lattice_vector_B3>{})".format(self.regexs.eol)),
                SM( " REAL SPACE MESH:\s+(?P<x_cpmd_real_space_mesh>{})".format(self.regexs.eol)),
                SM( " WAVEFUNCTION CUTOFF\(RYDBERG\):\s+(?P<x_cpmd_wave_function_cutoff>{})".format(self.regexs.float)),
                SM( " DENSITY CUTOFF\(RYDBERG\):          \(DUAL= 4\.00\)\s+(?P<x_cpmd_density_cutoff>{})".format(self.regexs.float)),
                SM( " NUMBER OF PLANE WAVES FOR WAVEFUNCTION CUTOFF:\s+(?P<x_cpmd_number_of_planewaves_wave_function>{})".format(self.regexs.int)),
                SM( " NUMBER OF PLANE WAVES FOR DENSITY CUTOFF:\s+(?P<x_cpmd_number_of_planewaves_density>{})".format(self.regexs.int)),
            ]
        )

    def initialization(self):
        """Returns the simplematcher that parser the atom initialization
        information. Common to all run types.
        """
        return SM(" GENERATE ATOMIC BASIS SET",
            forwardMatch=True,
            sections=["x_cpmd_section_wave_function_initialization"],
            subMatchers=[
            ]
        )

    def footer(self):
        """Returns the simplematcher that parser the CPMD footer containng the
        ending information. Common to all run types.
        """
        return SM( re.escape(" *                            TIMING                            *"),
            forwardMatch=True,
            subMatchers=[
                SM( re.escape(" *                            TIMING                            *"),
                    sections=["x_cpmd_section_timing"],
                    subMatchers=[
                    ]
                ),
                SM( "       CPU TIME :",
                    forwardMatch=True,
                    sections=["x_cpmd_section_end_information"],
                    subMatchers=[
                        # SM( " PROGRAM CPMD STARTED AT: (?P<x_cpmd_start_datetime>{})".format(self.regexs.eol)),
                    ]
                )
            ]
        )

    #===========================================================================
    # onOpen triggers
    def onOpen_section_method(self, backend, gIndex, section):
        self.cache_service["single_configuration_to_calculation_method_ref"] = gIndex

    def onOpen_section_system(self, backend, gIndex, section):
        self.cache_service["single_configuration_calculation_to_system_ref"] = gIndex

    #===========================================================================
    # onClose triggers
    def onClose_section_run(self, backend, gIndex, section):
        backend.addValue("program_name", "CPMD")
        backend.addValue("program_basis_set_type", "plane waves")

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        self.cache_service.addValue("single_configuration_calculation_to_system_ref")
        self.cache_service.addValue("single_configuration_to_calculation_method_ref")

    def onClose_section_system(self, backend, gIndex, section):
        self.cache_service.addArrayValues("configuration_periodic_dimensions")

    def onClose_section_method(self, backend, gIndex, section):
        backend.addValue("electronic_structure_method", "DFT")
        basis_id = backend.openSection("section_method_basis_set")
        backend.addValue("method_basis_set_kind", "wavefunction")
        backend.addValue("mapping_section_method_basis_set_cell_associated", 0)
        backend.closeSection("section_method_basis_set", basis_id)

    def onClose_x_cpmd_section_atom_kind(self, backend, gIndex, section):
        # Atomic kinds
        label = section.get_latest_value("x_cpmd_atom_kind_label")
        number = self.get_atom_number(label)
        id_kind = backend.openSection("section_method_atom_kind")
        backend.addValue("method_atom_kind_atom_number", number)
        backend.addValue("method_atom_kind_label", label)
        backend.closeSection("section_method_atom_kind", id_kind)

    def onClose_x_cpmd_section_start_information(self, backend, gIndex, section):
        # Starting date and time
        start_datetime = section.get_latest_value("x_cpmd_start_datetime")
        timestamp = self.timestamp_from_string(start_datetime)
        if timestamp:
            start_date_stamp, start_wall_time = timestamp
            backend.addValue("time_run_date_start", start_date_stamp)
            backend.addValue("time_run_wall_start", start_wall_time)

        # Input file
        input_filename = section.get_latest_value("x_cpmd_input_filename")
        input_filepath = self.file_service.set_file_id(input_filename, "input")
        if input_filepath is not None:
            input_parser = CPMDInputParser(self.parser_context)
            input_parser.parse(input_filepath)
        else:
            logger.warning("The input file for the calculation could not be found.")

        # Compilation date
        # date = section.get_latest_value("x_cpmd_compilation_date")

    def onClose_x_cpmd_section_supercell(self, backend, gIndex, section):
        # Simulation cell
        A1 = section.get_latest_value("x_cpmd_lattice_vector_A1")
        A2 = section.get_latest_value("x_cpmd_lattice_vector_A2")
        A3 = section.get_latest_value("x_cpmd_lattice_vector_A3")
        A1_array = self.vector_from_string(A1)
        A2_array = self.vector_from_string(A2)
        A3_array = self.vector_from_string(A3)
        cell = np.vstack((A1_array, A2_array, A3_array))
        self.cache_service["simulation_cell"] = cell

        # Plane wave basis
        cutoff = section.get_latest_value("x_cpmd_wave_function_cutoff")
        si_cutoff = convert_unit(cutoff, "rydberg")
        basis_id = backend.openSection("section_basis_set_cell_dependent")
        backend.addValue("basis_set_cell_dependent_kind", "plane_waves")
        backend.addValue("basis_set_cell_dependent_name", "PW_{}".format(cutoff))
        backend.addValue("basis_set_planewave_cutoff", si_cutoff)
        backend.closeSection("section_basis_set_cell_dependent", basis_id)

    def onClose_x_cpmd_section_run_type_information(self, backend, gIndex, section):
        geo_opt_method = section.get_latest_value("x_cpmd_geo_opt_method")
        geo_opt_method_mapping = {
            "GDIIS/BFGS": "bfgs",
            "LOW-MEMORY BFGS": "bfgs",
            "CONJUGATE GRADIENT": "conjugate_gradient",
            "STEEPEST DESCENT": "steepest_descent",
        }
        geo_opt_method = geo_opt_method_mapping.get(geo_opt_method)
        if geo_opt_method is not None:
            backend.addValue("geometry_optimization_method", geo_opt_method)

        # Number of steps
        n_steps = section.get_latest_value("x_cpmd_max_steps")
        self.cache_service["n_steps"] = n_steps

        # Temperature control for ions
        temp_control = section.get_latest_value("x_cpmd_ion_temperature_control")
        if temp_control is not None:
            if temp_control.strip() == "THE TEMPERATURE IS NOT CONTROLLED":
                self.cache_service["ensemble_type"] = "NVE"

        # Ions time step
        time_step_ions = section.get_latest_value("x_cpmd_time_step_ions")
        self.cache_service["time_step_ions"] = time_step_ions

    #===========================================================================
    # adHoc functions
    def ad_hoc_atom_information(self):
        """Parses the atom labels and coordinates.
        """
        def wrapper(parser):
            # Define the regex that extracts the information
            regex_string = r"\s+({0})\s+({1})\s+({2})\s+({2})\s+({2})\s+({0})".format(self.regexs.int, self.regexs.word, self.regexs.float)
            regex_compiled = re.compile(regex_string)

            match = True
            coordinates = []
            labels = []

            while match:
                line = parser.fIn.readline()
                result = regex_compiled.match(line)

                if result:
                    match = True
                    label = result.groups()[1]
                    labels.append(label)
                    coordinate = [float(x) for x in result.groups()[2:5]]
                    coordinates.append(coordinate)
                else:
                    match = False

            coordinates = np.array(coordinates)
            labels = np.array(labels)

            # If anything found, push the results to the correct section
            if len(coordinates) != 0:
                self.cache_service["initial_positions"] = coordinates
                self.cache_service["atom_labels"] = labels
                self.cache_service["number_of_atoms"] = coordinates.shape[0]

        return wrapper

    def ad_hoc_positions_forces(self, coord_metaname, force_metaname, coord_unit, force_unit):
        """Parses multiple lines with atom coordinates and forces.
        """
        def wrapper(parser):
            # Define the regex that extracts the information
            regex_string = r"\s+({0})\s+({1})\s+({2})\s+({2})\s+({2})\s+({2})\s+({2})\s+({2})".format(self.regexs.int, self.regexs.word, self.regexs.float)
            regex_compiled = re.compile(regex_string)

            match = True
            coordinates = []
            forces = []

            while match:
                line = parser.fIn.readline()
                result = regex_compiled.match(line)

                if result:
                    match = True
                    coordinate = [float(x) for x in result.groups()[2:5]]
                    force = [float(x) for x in result.groups()[5:8]]
                    coordinates.append(coordinate)
                    forces.append(force)
                else:
                    match = False

            coordinates = np.array(coordinates)
            forces = -np.array(forces)

            # If anything found, push the results to the correct section
            if len(coordinates) == len(forces) and len(coordinates) != 0:
                parser.backend.addArrayValues(coord_metaname, coordinates, unit=coord_unit)
                parser.backend.addArrayValues(force_metaname, forces, unit=force_unit)

        return wrapper

    def debug(self):
        def wrapper(parser):
            print("DEBUG")
        return wrapper

    #===========================================================================
    # Misc. functions
    def timestamp_from_string(self, timestring):

        class UTCtzinfo(datetime.tzinfo):
            """Class that represents the UTC timezone.
            """
            ZERO = datetime.timedelta(0)

            def utcoffset(self, dt):
                return UTCtzinfo.ZERO

            def tzname(self, dt):
                return "UTC"

            def dst(self, dt):
                return UTCtzinfo.ZERO

        """Returns the seconds since epoch for the given date and the wall
        clock seconds for the given wall clock time. Assumes UTC timezone.
        """
        # The start datetime can be formatted in different ways. First we try
        # to determine which way is used.
        try:
            timestring = timestring.strip()
            splitted = timestring.split()
            n_split = len(splitted)
            if n_split == 2:
                date, clock_time = timestring.split()
                year, month, day = [int(x) for x in date.split("-")]
                hour, minute, second, msec = [float(x) for x in re.split("[:.]", clock_time)]

                date_obj = datetime.datetime(year, month, day, tzinfo=UTCtzinfo())
                date_timestamp = (date_obj - datetime.datetime(1970, 1, 1, tzinfo=UTCtzinfo())).total_seconds()

                wall_time = hour*3600+minute*60+second+0.001*msec
            elif n_split == 5:
                weekday, month, day, clock_time, year = timestring.split()
                year = int(year)
                day = int(day)
                hour, minute, second = [float(x) for x in re.split("[:]", clock_time)]
                month_name_to_number = dict((v, k) for k, v in enumerate(calendar.month_abbr))
                month = month_name_to_number[month]

                date_obj = datetime.datetime(year, month, day, tzinfo=UTCtzinfo())
                date_timestamp = (date_obj - datetime.datetime(1970, 1, 1, tzinfo=UTCtzinfo())).total_seconds()
                wall_time = hour*3600+minute*60+second
        # If any sxception is encountered here, just return None as the
        # datetime is not very crucial
        except Exception:
            return None
        else:
            return date_timestamp, wall_time

    def get_atom_number(self, symbol):
        """ Returns the atomic number when given the atomic symbol.

        Args:
            symbol: atomic symbol as string

        Returns:
            The atomic number (number of protons) for the given symbol.
        """
        chemical_symbols = [
            'X',  'H',  'He', 'Li', 'Be',
            'B',  'C',  'N',  'O',  'F',
            'Ne', 'Na', 'Mg', 'Al', 'Si',
            'P',  'S',  'Cl', 'Ar', 'K',
            'Ca', 'Sc', 'Ti', 'V',  'Cr',
            'Mn', 'Fe', 'Co', 'Ni', 'Cu',
            'Zn', 'Ga', 'Ge', 'As', 'Se',
            'Br', 'Kr', 'Rb', 'Sr', 'Y',
            'Zr', 'Nb', 'Mo', 'Tc', 'Ru',
            'Rh', 'Pd', 'Ag', 'Cd', 'In',
            'Sn', 'Sb', 'Te', 'I',  'Xe',
            'Cs', 'Ba', 'La', 'Ce', 'Pr',
            'Nd', 'Pm', 'Sm', 'Eu', 'Gd',
            'Tb', 'Dy', 'Ho', 'Er', 'Tm',
            'Yb', 'Lu', 'Hf', 'Ta', 'W',
            'Re', 'Os', 'Ir', 'Pt', 'Au',
            'Hg', 'Tl', 'Pb', 'Bi', 'Po',
            'At', 'Rn', 'Fr', 'Ra', 'Ac',
            'Th', 'Pa', 'U',  'Np', 'Pu',
            'Am', 'Cm', 'Bk', 'Cf', 'Es',
            'Fm', 'Md', 'No', 'Lr'
        ]

        atom_numbers = {}
        for Z, name in enumerate(chemical_symbols):
            atom_numbers[name] = Z

        return atom_numbers[symbol]

    def vector_from_string(self, vectorstr):
        """Returns a numpy array from a string comprising of floating
        point numbers.
        """
        vectorstr = vectorstr.strip().split()
        vec_array = np.array([float(x) for x in vectorstr])
        return vec_array



# Copyright 2018-2018 Berk Onat
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from .GROMOSCommon import PARSERNAME, PROGRAMNAME, PARSERTAG
from nomadcore.smart_parser.SmartParserDictionary import metaNameConverter, metaNameConverter_UnderscoreSpaceDash
from nomadcore.smart_parser.SmartParserDictionary import MetaInfoMap, FileInfoMap, MapDictionary
from nomadcore.smart_parser.SmartParserDictionary import getDict_MetaStrInDict, getList_MetaStrInDict, get_unitDict
import nomadcore.md_data_access.MDDataAccess as MDData

def get_fileListDict():
    """Loads dictionary for file namelists.

    Returns:
        the list of defaults file namelists
    """

    startpage = {
        'nameTranslate'   : metaNameConverter,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_file',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_input_output_files'],
        'fileInterface'   : ["gromosread", "parmed", "mdtraj", "mdanalysis"]
        }
    namelist = {
        'structure'     : FileInfoMap(startpage, fileFormat=['.gromostop', '.top'], activeInfo=False,
                                      infoPurpose=['topology']),
        'input_coord'   : FileInfoMap(startpage, fileFormat=['.gromoscnf', '.cnf', '.g96'], activeInfo=False,
                                      infoPurpose=['topology', 'inputcoordinates', 'inputvelocities']),
        'output_coord'  : FileInfoMap(startpage, fileFormat=['.gromoscnf', '.cnf', '.g96'], activeInfo=False,
                                      infoPurpose=['outputcoordinates', 'outputvelocities']),
        'trajectory'    : FileInfoMap(startpage, fileFormat=['.gromostrc', '.trc'], activeInfo=False,
                                      infoPurpose=['trajectory', 'unitcell']),
        'traj_vel'      : FileInfoMap(startpage, fileFormat=['.gromostrv', '.trv'], activeInfo=False,
                                      infoPurpose=['velocities']),
        'traj_force'    : FileInfoMap(startpage, fileFormat=['.gromostrf', '.trf'], activeInfo=False,
                                      infoPurpose=['forces']),
        'output_log'    : FileInfoMap(startpage),
        'control_input' : FileInfoMap(startpage),
        }
    return MapDictionary(namelist)

def get_nameListDict(deflist):
    """Loads control in data of GROMOS.

    Args:
        deflist: name list definition (cntrl/parm).

    matchWith parameters:
        EOL = End of line
        PL  = Match with previous line
        PW  = Previous word
        NW  = Next word (until space, comma) (DEFAULT)
        UD  = until delimeter (can be any string/character)

        The text matched inside apostrophe or qoutation is extracted

    Returns:
        the list of namelists
    """
    startpage = {
        'nameTranslate'   : metaNameConverter_UnderscoreSpaceDash,
        'metaHeader'      : PARSERTAG,
        'metaNameTag'     : 'inout_control',
        'metaInfoType'    : 'C',
        'activeMetaNames' : [],
        'activeSections'  : [PARSERTAG+'_section_control_parameters'],
        }

    topocntrllist = {
        'MAKE_TOP' : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            replaceTag='topology_parameter_files', subFunc=lambda x:
            ', '.join(' '.join(x.strip().split()).replace(
                'topology,', '').replace(
                    'using:', '').replace('Force-field code:', ''
                        ).strip().split()[:-1]) if x is not None else None),
        'Force-field code' : MetaInfoMap(startpage, defaultValue=None,
                            subFunc=lambda x: 'Gromos-'+x if x is not None else None),
        'topology type' : MetaInfoMap(startpage, defaultValue=None,
                                      activeInfo=False, matchWith='PW'),
        'using atomistic' : MetaInfoMap(startpage, defaultValue=None,
                            replaceTag='topology_parameters', subFunc=lambda x:
                            'Atomistic' if x is not None else None),
        'using coarse-grained' : MetaInfoMap(startpage, defaultValue=None,
                                 replaceTag='topology_parameters', subFunc=lambda x:
                                 'Coarse-Grained' if x is not None else None),
        'RESNAME' : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            subFunc=lambda x: ', '.join(' '.join(x.strip().split()
                ).replace('END', '').replace(
                    'RESNAME', '').strip().split()
                ) if x is not None else None),
        'atom types' : MetaInfoMap(startpage, defaultValue=None, matchWith='PW',
            replaceTag='number_of_atom_types', matchAlso="ATOMTYPENAME",
            subFunc=lambda x: ' '.join(x.strip().split()
                ).replace('END', '').replace('ATOMTYPENAME', ''
                    ) if x is not None else None),
        'number of atoms' : MetaInfoMap(startpage, defaultValue=None, replaceTag='number of atoms solute',
            matchAlso="SOLUTEATOM", subFunc=lambda x: ' '.join(x.strip().split()
                ).replace('END', '').replace('SOLUTEATOM', ''
                    ) if x is not None else None),
        'Lennard-Jones exceptions' : MetaInfoMap(startpage, defaultValue=None, matchWith='PW',
            matchAlso="LJEXCEPTIONS", subFunc=lambda x: ' '.join(x.strip().split()
                ).replace('END', '').replace('LJEXCEPTIONS', ''
                    ) if x is not None else None),
        'bonds from BONDH block added to CONSTRAINT' : MetaInfoMap(startpage,
            defaultValue=None, matchWith='PW', replaceTag='number of H bonds at CONSTRAINT',
            matchAlso="END", subFunc=lambda x: ' '.join(x.strip().split()
                ).replace('END', '') if x is not None else None),
        'bonds from BOND block added to CONSTRAINT' : MetaInfoMap(startpage,
            defaultValue=None, matchWith='PW',
            replaceTag='number of bonds at CONSTRAINT', matchAlso="END"),
        'bondangles containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="BONDANGLE"),
        'bondangles not containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="BONDANGLE"),
        'improper dihedrals not containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="IMPDIHEDRAL"),
        'improper dihedrals containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="IMPDIHEDRAL"),
        'propdihedrals not containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="DIHEDRAL",
            replaceTag='dihedrals not containing hydrogens'),
        'propdihedrals containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="DIHEDRAL",
            replaceTag='dihedrals containing hydrogens'),
        'crossdihedrals not containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="CROSSDIHEDRAL"),
        'crossdihedrals containing hydrogens' : MetaInfoMap(startpage,
            defaultValue=None, matchAlso="CROSSDIHEDRAL"),
        'atoms' : MetaInfoMap(startpage, defaultValue=None, matchAlso="SOLVENT",
                replaceTag='number of solvent atoms'),
        'constraints' : MetaInfoMap(startpage, defaultValue=None, matchAlso="SOLVENT",
                replaceTag='number of solvent constraints'),
        'adding' : MetaInfoMap(startpage, defaultValue=None, matchAlso="SOLVENT",
                replaceTag='number of solvents added'),
        'molecules' : MetaInfoMap(startpage, defaultValue=None, matchAlso="SOLUTE",
                replaceTag='number of molecules in solute'),
        'temperature groups' : MetaInfoMap(startpage, defaultValue=None, matchAlso="SOLUTE",
                replaceTag='number of temperature groups for solute'),
        'pressure groups' : MetaInfoMap(startpage, defaultValue=None, matchAlso="SOLUTE",
                replaceTag='number of pressure groups for solute'),
        #''           : MetaInfoMap(startpage, defaultValue=None),
        }

    ffcntrllist = {
        'bond angle \(cosine\) interaction' : MetaInfoMap(startpage, defaultValue=None,
            matchAlso="FORCEFIELD", replaceTag='bond angle interaction in force field',
            subFunc=lambda x: 'Yes' if x is not None else None),
        'improper-dihedral-interaction' : MetaInfoMap(startpage, defaultValue=None,
            matchAlso="FORCEFIELD", replaceTag='improper dihedral interaction in force field',
            subFunc=lambda x: 'Yes' if x is not None else None),
        'propdihedral_interaction' : MetaInfoMap(startpage, defaultValue=None,
            matchAlso="FORCEFIELD", replaceTag='dihedral interaction in force field',
            subFunc=lambda x: 'Yes' if x is not None else None),
        'crossdihedral interaction' : MetaInfoMap(startpage, defaultValue=None,
            matchAlso="FORCEFIELD", replaceTag='cross dihedral interaction in force field',
            subFunc=lambda x: 'Yes' if x is not None else None),
        'nonbonded force' : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            matchAlso="FORCEFIELD", replaceTag='nonbonded definitions in force field',
            subFunc=lambda x: ', '.join(x.strip().split()) if x is not None else None),
        'Pairlist Algorithm'  : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            subFunc=lambda x: ' '.join(x.strip().split()) if x is not None else None),
        'boundary'  : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            replaceTag='periodic boundary conditions',
            subFunc=lambda x: ' '.join(x.strip().split()) if x is not None else None),
        'virial'  : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL',
            subFunc=lambda x: ' '.join(x.strip().split()) if x is not None else None),
        'cutoff '  : MetaInfoMap(startpage, defaultValue=None, matchAlso='pairlist creation',
                                  replaceTag='cutoff type'),
        'shortrange-cutoff-'  : MetaInfoMap(startpage, defaultValue=None,
                                      replaceTag='shortrange cutoff'),
        'longrange-cutoff-'  : MetaInfoMap(startpage, defaultValue=None,
                                      replaceTag='longrange cutoff'),
        'pairlist creation every'  : MetaInfoMap(startpage, defaultValue=None,
                                      replaceTag='pairlist update step frequency'),
        'reactionfield cutoff '  : MetaInfoMap(startpage, defaultValue=None,
            replaceTag='reactionfield cutoff'),
        'epsilon '  : MetaInfoMap(startpage, defaultValue=None,
            replaceTag='force field epsilon'),
        'reactionfield-epsilon-'  : MetaInfoMap(startpage, defaultValue=None,
            replaceTag='reactionfield epsilon'),
        'kappa'  : MetaInfoMap(startpage, defaultValue=None,
            replaceTag='force field kappa'),
        'perturbation'  : MetaInfoMap(startpage, defaultValue=None,
            replaceTag='force field perturbation'),
        }

    filecntrllist = {
        'parameter read from'  : MetaInfoMap(startpage, defaultValue=None,
                                             replaceTag='input file'),
        'topology read from'  : MetaInfoMap(startpage, defaultValue=None,
                                             replaceTag='topology file'),
        'configuration read from'  : MetaInfoMap(startpage, defaultValue=None,
                                             replaceTag='configuration file'),
        }

    cntrllist = {
        'TITLE'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-NTEM'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-NCYC'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-DELE'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-DX0'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-DXM'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-NMIN'  : MetaInfoMap(startpage, defaultValue=None),
        'EMIN-FLIM'  : MetaInfoMap(startpage, defaultValue=None),
        'SYS-NPM'  : MetaInfoMap(startpage, defaultValue=None),
        'SYS-NSM'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTIVEL'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTISHK'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTINHT'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTINHB'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTISHI'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTIRTC'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTICOM'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-NTISTI'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-IG'  : MetaInfoMap(startpage, defaultValue=None),
        'INIT-TEMPI'  : MetaInfoMap(startpage, defaultValue=None),
        'STEP-NSTLIM'  : MetaInfoMap(startpage, defaultValue=None),
        'STEP-T'  : MetaInfoMap(startpage, defaultValue=None),
        'STEP-DT'  : MetaInfoMap(startpage, defaultValue=None),
        'BCND-NTB'  : MetaInfoMap(startpage, defaultValue=None),
        'BCND-NDFMIN'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-ALG'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-NUM'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-NBATHS'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-TEMP'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-TAU'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-DOFSET'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-LAST'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-COMBATH'  : MetaInfoMap(startpage, defaultValue=None),
        'BATH-IRBATH'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-COUPLE'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-SCALE'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-COMP'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-TAUP'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-VIRIAL'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-ANISO'  : MetaInfoMap(startpage, defaultValue=None),
        'PRES-INIT0'  : MetaInfoMap(startpage, defaultValue=None),
        'COVF-NTBBH'  : MetaInfoMap(startpage, defaultValue=None),
        'COVF-NTBAH'  : MetaInfoMap(startpage, defaultValue=None),
        'COVF-NTBDN'  : MetaInfoMap(startpage, defaultValue=None),
        'SOLM-NSPM'  : MetaInfoMap(startpage, defaultValue=None),
        'SOLM-NSP'  : MetaInfoMap(startpage, defaultValue=None),
        'COMT-NSCM'  : MetaInfoMap(startpage, defaultValue=None),
        'PRNT-NTPR'  : MetaInfoMap(startpage, defaultValue=None),
        'PRNT-NTPP'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWX'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWSE'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWV'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWF'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWE'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWG'  : MetaInfoMap(startpage, defaultValue=None),
        'WRIT-NTWB'  : MetaInfoMap(startpage, defaultValue=None),
        'CNST-NTC'  : MetaInfoMap(startpage, defaultValue=None),
        'CNST-NTCP'  : MetaInfoMap(startpage, defaultValue=None),
        'CNST-NTCP0'  : MetaInfoMap(startpage, defaultValue=None),
        'CNST-NTCS'  : MetaInfoMap(startpage, defaultValue=None),
        'CNST-NTCS0'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-BONDS'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-ANGS'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-IMPS'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-DIHS'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-ELEC'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-VDW'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-NEGR'  : MetaInfoMap(startpage, defaultValue=None),
        'FORC-NRE'  : MetaInfoMap(startpage, defaultValue=None),
        'PAIR-ALG'  : MetaInfoMap(startpage, defaultValue=None),
        'PAIR-NSNB'  : MetaInfoMap(startpage, defaultValue=None),
        'PAIR-RCUTP'  : MetaInfoMap(startpage, defaultValue=None),
        'PAIR-RCUTL'  : MetaInfoMap(startpage, defaultValue=None),
        'PAIR-SIZE'  : MetaInfoMap(startpage, defaultValue=None),
        'PAIR-TYPE'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NLRELE'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-APPAK'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-RCRF'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-EPSRF'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NSLFEXCL'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NSHAPE'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-ASHAPE'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NA2CLC'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-TOLA2'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-EPSLS'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NKX'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NKY'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NKZ'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-KCUT'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NGX'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NGY'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NGZ'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NASORD'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NFDORD'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NALIAS'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NSPORD'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NQEVAL'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-FACCUR'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NRDGRD'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NWRGRD'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-NLRLJ'  : MetaInfoMap(startpage, defaultValue=None),
        'NONB-SLVDNS'  : MetaInfoMap(startpage, defaultValue=None),
        'Parameter' : MetaInfoMap(startpage, defaultValue=None, matchWith='EOL'),
        }

    startpage.update({
        'metaNameTag'     : 'mdout',
        'activeSections'  : ['section_single_configuration_calculation']
        })
    mddatalist = {
        'STEP'      : MetaInfoMap(startpage, matchWith='EOL',
            subFunc=lambda x: x.strip().split()[1] if len(x.strip().split())>1 is not None else None),
        'TIME'      : MetaInfoMap(startpage, matchWith='EOL',
            subFunc=lambda x: x.strip().split()[1] if len(x.strip().split())>1 is not None else None),
        'T_avg'   : MetaInfoMap(startpage, matchWith='EOL',
            subFunc=lambda x: x.strip().split()[4] if len(x.strip().split())>1 is not None else None),
        'E_Total'   : MetaInfoMap(startpage, matchAlso=":"),
        'E_Kinetic' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Potential\+Special' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_Potential_Special'),
        'E_Potential' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Covalent' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Bonds'   : MetaInfoMap(startpage, matchAlso=":"),
        'E_Angles'  : MetaInfoMap(startpage, matchAlso=":"),
        'E_Improper' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Dihedral' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Crossdihedral' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Non-bonded' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Vdw'     : MetaInfoMap(startpage, matchAlso=":"),
        'E_El \(RF\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_RF'),
        'E_El \(LF\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_LF'),
        'E_El \(pair\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_pair'),
        'E_El \(real-space\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_real-space'),
        'E_El \(k-space\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_k-space'),
        'E_El \(A term\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_A term'),
        'E_El \(lattice sum self\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_lattice sum self'),
        'E_El \(surface term\)' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_El_surface term'),
        'E_Polarisation self' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Special' : MetaInfoMap(startpage, matchAlso=":"),
        'E_SASA                 :' : MetaInfoMap(startpage,
            replaceTag='E_SASA'),
        'E_SASA Volume' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Constraints' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Distanceres' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Disfieldres' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Dihrest' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Posrest' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Jrest'   : MetaInfoMap(startpage, matchAlso=":"),
        'E_X-ray restraints' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Local elevation' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Order-parameter rest.' : MetaInfoMap(startpage, matchAlso=":",
            replaceTag='E_Order-parameter-rest'),
        'E_RDCrest' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Symmetry restraints' : MetaInfoMap(startpage, matchAlso=":"),
        'E_EDS reference' : MetaInfoMap(startpage, matchAlso=":"),
        'E_Entropy' : MetaInfoMap(startpage, matchAlso=":"),
        'E_QM'      : MetaInfoMap(startpage, matchAlso=":"),
        'pressure'  : MetaInfoMap(startpage, matchAlso=":"),
        'volume'    : MetaInfoMap(startpage, matchAlso=":"),
        'virial'    : MetaInfoMap(startpage),
        #'molecular kinetic energy'    : MetaInfoMap(startpage),
        'pressure tensor'    : MetaInfoMap(startpage),
        }

    startpage.update({
        'metaNameTag'     : 'parm',
        'activeSections'  : [PARSERTAG+'_mdin_method']
        })
    extralist = {
        'flags' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage),
        'minimization' : MetaInfoMap(startpage, activeInfo=True),
        'integrator' : MetaInfoMap(startpage, activeInfo=True),
        'thermostat' : MetaInfoMap(startpage, activeInfo=True),
        'barostat' : MetaInfoMap(startpage, activeInfo=True),
        'box_info' : MetaInfoMap(startpage),
        'unitcell_radius' : MetaInfoMap(startpage),
        'total_memory' : MetaInfoMap(startpage),
        'file_format' : MetaInfoMap(startpage),
        'file_version' : MetaInfoMap(startpage),
        'file_date' : MetaInfoMap(startpage),
        'file_time' : MetaInfoMap(startpage)
        }

    if deflist == 'mddata':
        namelist = mddatalist
    elif deflist == 'topocntrl':
        namelist = topocntrllist
    elif deflist == 'ffcntrl':
        namelist = ffcntrllist
    elif deflist == 'extra':
        namelist = extralist
    elif deflist == 'filecntrl':
        namelist = filecntrllist
    elif deflist == 'stepcontrol':
        namelist = stepcontrol
    else:
        namelist = cntrllist
    return MapDictionary(namelist)

def set_Dictionaries(self):
    self.unitDict = get_unitDict('si')
    self.fileDict = get_fileListDict()
    self.cntrlDict = get_nameListDict('cntrl')
    self.topocntrlDict = get_nameListDict('topocntrl')
    self.ffcntrlDict = get_nameListDict('ffcntrl')
    self.filecntrlDict = get_nameListDict('filecntrl')
    #self.parmDict = get_nameListDict('parm')
    self.mddataDict = get_nameListDict('mddata')
    self.extraDict = get_nameListDict('extra')
    self.mdoutHeaderDict = {}
    #self.fileUnitDict = {}
    self.stepcontrolDict = {
            'logsteps'       : None,
            'nextlogsteps'   : None,
            'trajsteps'      : None,
            'velsteps'       : None,
            'forcesteps'     : None,
            'steps'          : None,
            'follow'         : None,
            'sectioncontrol' : None,
            'cntrlparmstep'  : [-1],
        }
    self.metaDicts = {
            'file'   : self.fileDict,
            'cntrl'  : self.cntrlDict,
            'topocntrl'  : self.topocntrlDict,
            'ffcntrl'  : self.ffcntrlDict,
            'filecntrl'  : self.filecntrlDict,
#            'parm'   : self.parmDict,
            'mddata' : self.mddataDict,
            'extra'  : self.extraDict,
            }
    self.sectionDict = {
            'section' : {
                "metaNameTag"      : ['input_output_files',
                                      'control_parameters'],
                "listTypStr"       : 'type_section',
                "repeatingSection" : False,
                "supraNames"       : ["section_run"]
                }
            }

def get_updateDictionary(self, defname):

    def value_convert_array(itemdict):
        val = itemdict["value"]
        if(val is not None and isinstance(val, str)):
            if("[" in val or "(" in val):
                return False, np.asarray(val), itemdict
        return False, val, itemdict

    def value_convert_array_norm(itemdict):
        val = []
        depdict = itemdict["depends"]
        for item in depdict["value"]:
            st=self.metaStorage.fetchAttrValue(item)
            if(st is not None and isinstance(st, str)):
                if("[" in st or "(" in st):
                    val.append(list(storedtext))
        if val:
            return False, np.linalg.norm(np.asarray(val)), itemdict
        else:
            return False, itemdict["value"], itemdict

    startpage = {
        'metaHeader'      : '',
        'metaNameTag'     : '',
        'activeMetaNames' : [],
        'activeSections'  : ['section_sampling_method','section_single_configuration_calculation',PARSERTAG+'_mdout']
        }

    # ---------------------------------------------------------------
    #   Definitions of meta data values for section_sampling_method
    # ---------------------------------------------------------------
    sampling = {
        #'ensemble_type' : MetaInfoMap(startpage, activeInfo=True,
        #    depends=[
        #        {'test' : [['DYNA', ' is not None'],
        #                   ['FIRSTT', ' in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\",'
        #                                  '\"langevin\"]'],
        #                   ['barostat', ' in [\"berendsen\",'
        #                                 '\"langevin\"]']],
        #         'assign' : 'NPT'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' in [\"langevin\"]'],
        #                   ['thermostat', ' not in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\"]'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'Langevin'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' in [\"tcouple\",'
        #                                  '\"rescale\",'
        #                                  '\"reassign\",'
        #                                  '\"andersen\"]'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'NVT'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' is None'],
        #                   ['barostat', ' is None']],
        #         'assign' : 'NVE'},
        #        {'test' : [['minimization', ' is not None']],
        #         'assign' : 'minimization'},
        #        ],
        #    lookupdict=self.extraDict
        #    ),
        'sampling_method' : MetaInfoMap(startpage, activeInfo=True,
            depends=[
                {'test' : [['MINI', ' is not None']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['DYNA', ' is not None']],
                 'assign' : 'molecular_dynamics'}
                ],
            lookupdict=self.cntrlDict
            ),
#            'settings_geometry_optimization' : MetaInfoMap(startpage),
#            'settings_metadynamics' : MetaInfoMap(startpage),
#            'settings_molecular_dynamics' : MetaInfoMap(startpage),
#            'settings_Monte_Carlo' : MetaInfoMap(startpage),
#        'geometry_optimization_energy_change' : MetaInfoMap(startpage,
#            depends={
#                '' : {'imin' : '1'},
#                },
#            lookupdict=self.cntrlDict
#            ),
#       'geometry_optimization_geometry_change' : MetaInfoMap(startpage),
        'geometry_optimization_method' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"SD\",\"sd\"]']],
                 'assign' : 'SD'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"ABNR\",\"abnr\"]']],
                 'assign' : 'Adopted Basis Newton-Raphson'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"NRAP\",\"nrap\"]']],
                 'assign' : 'Newton-Raphson'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"POWELL\",\"powell\"]']],
                 'assign' : 'Powell'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"POWE\",\"powe\"]']],
                 'assign' : 'Powell'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' in [\"TNPACK\",\"tnpack\"]']],
                 'assign' : 'Truncated Newton Method'},
                {'test' : [['MINI',' is not None'],
                           ['MINI', ' not in [\"SD\",\"sd\",'
                                             '\"ABNR\",\"abnr\",'
                                             '\"NRAP\",\"nrap\",'
                                             '\"POWE\",\"powe\",'
                                             '\"POWELL\",\"powell\",'
                                             '\"TNPACK\",\"tnpack\"]']],
                 'assign' : 'CG'},
                ],
            lookupdict=self.cntrlDict,
            activeInfo=True,
            ),
#       'geometry_optimization_threshold_force' : MetaInfoMap(startpage),
        #'x_gromos_barostat_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['minimization', ' is None'],
        #                   ['barostat', '== \"berendsen\"']],
        #         'assign' : 'Berendsen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['barostat', '== \"langevin\"']],
        #         'assign' : 'Nose-Hoover Langevin'}
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_gromos_barostat_target_pressure' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['TARGET PRESSURE', ' is not None']],
        #         'value' : 'TARGET PRESSURE'}
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='atmosphere',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_gromos_barostat_tau' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['LANGEVIN OSCILLATION PERIOD', ' is not None']],
        #         'value' : 'LANGEVIN OSCILLATION PERIOD'},
        #        {'test' : [['BERENDSEN RELAXATION TIME', ' is not None']],
        #         'value' : 'BERENDSEN RELAXATION TIME'}
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='akmatime',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_barostat']
        #    ),
        #'x_gromos_integrator_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['integrator', ' is not None']],
        #         'value' : 'integrator'},
        #        {'test' : [['thermostat', ' == \"langevin\"']],
        #         'assign' : 'Langevin'}
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_integrator']
        #    ),
        'x_gromos_integrator_dt' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['DYNA', ' is not None'],
                           ['TIME STEP', ' is not None']],
                 'value' : 'TIME STEP'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='nano-seconds',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        'x_gromos_number_of_steps_requested' : MetaInfoMap(startpage,
            depends=[{'value' : 'NSTEP'}],
            lookupdict=self.cntrlDict,
            valtype='int',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_integrator']
            ),
        #'x_gromos_thermostat_type' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"tcouple\"']],
        #         'assign' : 'Berendsen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"rescale\"']],
        #         'assign' : 'Velocity Rescaling'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"reassign\"']],
        #         'assign' : 'Velocity Reassigning'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"andersen\"']],
        #         'assign' : 'Andersen'},
        #        {'test' : [['minimization', ' is None'],
        #                   ['thermostat', ' == \"langevin\"']],
        #         'assign' : 'Nose-Hoover Langevin'},
        #        ],
        #    lookupdict=self.extraDict,
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_thermostat']
        #    ),
        'x_gromos_thermostat_target_temperature' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['DYNA', ' is not None'],
                           ['FINALT', ' is not None']],
                 'value' : 'FINALT'},
                ],
            lookupdict=self.cntrlDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            ),
        #'x_gromos_thermostat_tau' : MetaInfoMap(startpage,
        #    depends=[
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['VELOCITY RESCALE FREQ', ' is not None']],
        #         'value' : 'VELOCITY RESCALE FREQ'},
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['VELOCITY REASSIGNMENT FREQ', ' is not None']],
        #         'value' : 'VELOCITY REASSIGNMENT FREQ'},
        #        {'test' : [['MINIMIZATION', ' is None'],
        #                   ['VELOCITY QUENCHING', ' is None'],
        #                   ['LOWE-ANDERSEN RATE', ' is not None']],
        #         'value' : 'LOWE-ANDERSEN RATE'},
        #        ],
        #    lookupdict=self.cntrlDict,
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='pico-second',
        #    #autoSections=True,
        #    activeInfo=True,
        #    activeSections=['settings_thermostat']
        #    ),
        'x_gromos_periodicity_type' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['CRYSTAL', ' is not None']],
                 'assign' : 'periodic boundaries'},
                {'test' : [['CRYSTAL',' is None']],
                 'assign' : 'no periodic boundaries'},
                ],
            lookupdict=self.cntrlDict,
            #autoSections=True,
            activeInfo=True,
            activeSections=['settings_thermostat']
            )
        }

    # ------------------------------------------------------------
    #   Definitions for section_single_configuration_calculation
    # ------------------------------------------------------------
    singleconfcalc = {
        #'atom_forces_type' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_correction_entropy' : MetaInfoMap(startpage),
        'energy_current' : MetaInfoMap(startpage),
        'energy_electrostatic' : MetaInfoMap(startpage,
            depends=[{'value' : 'E_Non-bonded'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'energy_free_per_atom' : MetaInfoMap(startpage),
        'energy_free' : MetaInfoMap(startpage),
        #'energy_method_current' : MetaInfoMap(startpage,
        #    depends=[{'assign' : 'Force Field'}],
        #    lookupdict=self.mddataDict
        #    ),
        'energy_T0_per_atom' : MetaInfoMap(startpage),
        'energy_total_T0_per_atom' : MetaInfoMap(startpage),
        'energy_total_T0' : MetaInfoMap(startpage,
            depends=[{'value' : 'E_Potential'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'energy_total' : MetaInfoMap(startpage,
            depends=[{'value' : 'E_Total'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            activeInfo=True,
            lookupdict=self.mddataDict
            ),
        'hessian_matrix' : MetaInfoMap(startpage),
        'single_configuration_calculation_converged' : MetaInfoMap(startpage),
        'single_configuration_calculation_to_system_ref' : MetaInfoMap(startpage),
        'single_configuration_to_calculation_method_ref' : MetaInfoMap(startpage),
        'time_calculation' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_cpu1_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_date_start' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_end' : MetaInfoMap(startpage),
        'time_single_configuration_calculation_wall_start' : MetaInfoMap(startpage),
        'stress_tensor_kind' : MetaInfoMap(startpage,
            depends=[
                {'test' : [['MINI', ' is not None']],
                 'assign' : 'geometry_optimization'},
                {'test' : [['DYNA', ' is not None']],
                 'assign' : 'molecular_dynamics'}
                ],
            activeInfo=True,
            lookupdict=self.extraDict
            ),
        'stress_tensor_value' : MetaInfoMap(startpage)
        }

    # section_single_energy_van_der_Waals of section_single_configuration_calculation
    singlevdw = {
        'energy_van_der_Waals_value' : MetaInfoMap(startpage,
            depends=[{'value' : 'E_Vdw'}],
            lookupdict=self.mddataDict,
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            #autoSections=True,
            activeSections=['section_energy_van_der_Waals']
            ),
        }

    # Info for section_restricted_uri
    restrictions = {
        'number_of_restricted_uri_files' : MetaInfoMap(startpage,
            value=int(1),
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_files' : MetaInfoMap(startpage,
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri'],
            subfunction={
                'function' : self.parameter_file_name,
                'supportDict' : self.fileDict,
                },
            ),
        'restricted_uri_license' : MetaInfoMap(startpage,
            value='GROMOS License',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_restriction' : MetaInfoMap(startpage,
            value='any access',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_reason' : MetaInfoMap(startpage,
            value='propriety license',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        'restricted_uri_issue_authority' : MetaInfoMap(startpage,
            value='GROMOS',
            #autoSections=True,
            activeInfo=True,
            activeSections=['section_restricted_uri']
            ),
        }

    # ------------------------------------------
    #   Definitions for section_frame_sequence
    # ------------------------------------------
    frameseq = {
        #'frame_sequence_conserved_quantity_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TS'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_conserved_quantity_stats' : MetaInfoMap(startpage),
        #'frame_sequence_conserved_quantity' : MetaInfoMap(startpage,
        #    depends=[{'store' : ''}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_continuation_kind' : MetaInfoMap(startpage),
        'frame_sequence_external_url' : MetaInfoMap(startpage),
        'frame_sequence_kinetic_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_kinetic_energy_stats' : MetaInfoMap(startpage),
        'frame_sequence_kinetic_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'E_Kinetic'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_local_frames_ref' : MetaInfoMap(startpage),
        'frame_sequence_potential_energy_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_potential_energy_stats' : MetaInfoMap(startpage),
        'frame_sequence_potential_energy' : MetaInfoMap(startpage,
            depends=[{'store' : 'E_Potential'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_pressure_average_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_pressure_average' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'PRSI'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='atmosphere',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_pressure_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_pressure_stats' : MetaInfoMap(startpage),
        'frame_sequence_pressure' : MetaInfoMap(startpage,
            depends=[{'store' : 'Pressure'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/(mol-(nano-meter)**3)',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_temperature_average_frames' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TS'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'frame_sequence_temperature_average' : MetaInfoMap(startpage,
        #    depends=[{'store' : 'TEMPAVG'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='Kelvin',
        #    lookupdict=self.mddataDict
        #    ),
        'frame_sequence_temperature_frames' : MetaInfoMap(startpage,
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_temperature_stats' : MetaInfoMap(startpage),
        'frame_sequence_temperature' : MetaInfoMap(startpage,
            depends=[{'store' : 'T_avg'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='Kelvin',
            lookupdict=self.mddataDict
            ),
        'frame_sequence_time' : MetaInfoMap(startpage,
            depends=[{'store' : 'TIME'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='nano-second',
            # Calculates in AKMA units including inputs but prints in picoseconds.
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_volume_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_volume' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'volume'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='(nano-meter)**3',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_total_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_total_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Total'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_total_kinetic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_total_kinetic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Kinetic'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        #'x_gromos_frame_sequence_virial_kinetic_energy_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_gromos_frame_sequence_virial_kinetic_energy_energy' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'VIRKE'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit=self.unitDict['kilo-joule/mol'],
        #    lookupdict=self.mddataDict
        #    ),
        #'x_gromos_frame_sequence_boundary_frames' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'STEP'}],
        #    valtype='int',
        #    lookupdict=self.mddataDict
        #    ),
        #'x_gromos_frame_sequence_boundary' : MetaInfoMap(startpage,
        #    activeSections=['section_frame_sequence'],
        #    depends=[{'store' : 'BOUNDARY'}],
        #    valtype='float',
        #    unitdict=self.unitDict,
        #    unit='kilo-joule/mol',
        #    lookupdict=self.mddataDict
        #    ),
        'x_gromos_frame_sequence_bond_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_bond_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Bonds'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_angle_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_angle_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Angles'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_proper_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_proper_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Dihedral'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_improper_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_improper_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Improper'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_cross_dihedral_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_cross_dihedral_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Crossdihedral'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_vdw_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_vdw_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Vdw'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_electrostatic_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_electrostatic_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Non-bonded'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_covalent_energy_frames' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'STEP'}],
            valtype='int',
            lookupdict=self.mddataDict
            ),
        'x_gromos_frame_sequence_covalent_energy' : MetaInfoMap(startpage,
            activeSections=['section_frame_sequence'],
            depends=[{'store' : 'E_Covalent'}],
            valtype='float',
            unitdict=self.unitDict,
            unit='kilo-joule/mol',
            lookupdict=self.mddataDict
            ),
        #'frame_sequence_to_sampling_ref' : MetaInfoMap(startpage),
        'geometry_optimization_converged' : MetaInfoMap(startpage,
            depends=[
                {'test' : [["MINI", " is not None"],
                           ['Minimization exiting with', ' is not None']],
                 'assign' : True},
                ],
            activeInfo=True,
            lookupdict=self.cntrlDict
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    frameseqend = {
        #'number_of_conserved_quantity_evaluations_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'frame_sequence_conserved_quantity_frames'
        #            )
        #        )
        #    ),
        'number_of_frames_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_kinetic_energies_in_sequence' : MetaInfoMap(startpage, activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_kinetic_energy_frames'
                    )
                )
            ),
        'number_of_potential_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_potential_energy_frames'
                    )
                )
            ),
        'number_of_pressure_evaluations_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_pressure_frames'
                    )
                )
            ),
        'number_of_temperatures_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'frame_sequence_temperature_frames'
                    )
                )
            ),
        'x_gromos_number_of_volumes_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_volume_frames'
                    )
                )
            ),
        #'x_gromos_number_of_boundary_energies_in_sequence' : MetaInfoMap(startpage,
        #    activeInfo=True,
        #    activeSections=['section_frame_sequence'],
        #    value=(lambda x: np.array(x['val']).flatten().shape[0] if(
        #        x is not None and x['val'] is not None) else None)(
        #        self.metaStorage.fetchAttrValue(
        #            'x_gromos_frame_sequence_boundary_frames'
        #            )
        #        )
        #    ),
        'x_gromos_number_of_bond_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_bond_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_angle_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_angle_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_electrostatic_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_electrostatic_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_vdw_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_vdw_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_proper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_proper_dihedral_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_improper_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_improper_dihedral_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_cross_dihedral_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_cross_dihedral_energy_frames'
                    )
                )
            ),
        'x_gromos_number_of_covalent_energies_in_sequence' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_frame_sequence'],
            value=(lambda x: np.array(x['val']).flatten().shape[0] if(
                x is not None and x['val'] is not None) else None)(
                self.metaStorage.fetchAttrValue(
                    'x_gromos_frame_sequence_covalent_energy_frames'
                    )
                )
            ),
        #'previous_sequence_ref' : MetaInfoMap(startpage)
        }

    # ----------------------------------
    #   Definitions for section_system
    # ----------------------------------
    # section_system
    sec_system = {
        #'topology_ref' : MetaInfoMap(startpage),
        'atom_velocities' : MetaInfoMap(startpage),
        'configuration_raw_gid' : MetaInfoMap(startpage),
        'local_rotations' : MetaInfoMap(startpage),
        'number_of_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_sites' : MetaInfoMap(startpage),
        'number_of_symmetry_operations' : MetaInfoMap(startpage),
        'reduced_symmetry_matrices' : MetaInfoMap(startpage),
        'reduced_symmetry_translations' : MetaInfoMap(startpage),
        'SC_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_choice' : MetaInfoMap(startpage),
        'spacegroup_3D_hall' : MetaInfoMap(startpage),
        'spacegroup_3D_international' : MetaInfoMap(startpage),
        'spacegroup_3D_number' : MetaInfoMap(startpage),
        'spacegroup_3D_origin_shift' : MetaInfoMap(startpage),
        'spacegroup_3D_pointgroup' : MetaInfoMap(startpage),
        'spacegroup_3D_std_lattice' : MetaInfoMap(startpage),
        'spacegroup_3D_std_positions' : MetaInfoMap(startpage),
        'spacegroup_3D_std_types' : MetaInfoMap(startpage),
        'spacegroup_3D_trasformation_matrix' : MetaInfoMap(startpage),
        'spacegroup_3D_wyckoff' : MetaInfoMap(startpage),
        'symmorphic' : MetaInfoMap(startpage),
        'system_name' : MetaInfoMap(startpage,
            activeInfo=True,
            activeSections=['section_system'],
            subfunction={
                'function' : MDData.MDDataConverter.topology_system_name,
                'supportDict' : self.topoDict,
                },
            #functionbase=self
            ),
        'time_reversal_symmetry' : MetaInfoMap(startpage)
        }

    # section_configuration_core of section_system
    configuration_core = {
        'number_of_electrons' : MetaInfoMap(startpage,
            #value=0,
            #valtype='float',
            ),
        'atom_labels' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_labels()
            ),
        'atom_positions' : MetaInfoMap(startpage,
            #subfunction=self.system_atom_positions()
            ),
        'configuration_periodic_dimensions' : MetaInfoMap(startpage,
            ),
        'embedded_system' : MetaInfoMap(startpage),
        'lattice_vectors' : MetaInfoMap(startpage,
            #subfunction=self.system_lattice_vectors()
            ),
        'simulation_cell' : MetaInfoMap(startpage,
            #subfunction=self.system_simulation_cell()
            )
        }

    # section_spacegroup_3D_operation of section_system
    spacegroup_op = {
            'spacegroup_3D_rotation' : MetaInfoMap(startpage),
            'spacegroup_3D_translation' : MetaInfoMap(startpage)
        }

    # section_system_to_system_refs of section_system
    sys_to_sys = {
            'system_to_system_kind' : MetaInfoMap(startpage),
            'system_to_system_ref' : MetaInfoMap(startpage)
        }

    # --------------------------------------------------------
    #   Definitions of meta data values for section_topology
    # --------------------------------------------------------
    # section_topology of section_run
    topology = {
        'atom_to_molecule' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_atom_to_mol,
                'supportDict' : self.topoDict,
                },
            #functionbase=self
            ),
        'molecule_to_molecule_type_map' : MetaInfoMap(startpage),
        'number_of_topology_atoms' : MetaInfoMap(startpage,
            depends=[{'value' : 'number_of_atoms'}],
            valtype='int',
            lookupdict=self.extraDict
            ),
        'number_of_topology_molecules' : MetaInfoMap(startpage,
            subfunction={
                'function' : MDData.MDDataConverter.topology_num_topo_mol,
                'supportDict' : self.topoDict,
                },
            valtype='int',
            ),
        'topology_force_field_name' : MetaInfoMap(startpage,
            depends=[{'value' : 'PARAMETER file'}],
            lookupdict=self.cntrlDict
            )
        }

    # section_atom_type of section_topology
    atom_type = {
        'atom_type_charge' : MetaInfoMap(startpage),
        'atom_type_mass' : MetaInfoMap(startpage),
        'atom_type_name' : MetaInfoMap(startpage)
        }

    # section_constraint of section_topology
    constraint = {
        'constraint_atoms' : MetaInfoMap(startpage),
        'constraint_kind' : MetaInfoMap(startpage),
        'constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_constraint' : MetaInfoMap(startpage),
        'number_of_constraints' : MetaInfoMap(startpage)
        }

    # section_interaction of section_topology
    interaction = {
        'interaction_atoms' : MetaInfoMap(startpage),
        'interaction_kind' : MetaInfoMap(startpage),
        'interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_interaction' : MetaInfoMap(startpage),
        'number_of_interactions' : MetaInfoMap(startpage)
        }

    # -------------------------------------------------------------
    #   Definitions of meta data values for section_molecule_type
    # -------------------------------------------------------------
    # section_molecule_type of section_topology
    mol_type = {
        'molecule_type_name' : MetaInfoMap(startpage),
        'number_of_atoms_in_molecule' : MetaInfoMap(startpage),
        'settings_atom_in_molecule' : MetaInfoMap(startpage)
        }

    # section_molecule_constraint of section_molecule_type
    mol_constraint = {
        'molecule_constraint_atoms' : MetaInfoMap(startpage),
        'molecule_constraint_kind' : MetaInfoMap(startpage),
        'molecule_constraint_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_constraint' : MetaInfoMap(startpage),
        'number_of_molecule_constraints' : MetaInfoMap(startpage)
        }

    # section_molecule_interaction of section_molecule_type
    mol_interaction = {
        'molecule_interaction_atoms' : MetaInfoMap(startpage),
        'molecule_interaction_kind' : MetaInfoMap(startpage),
        'molecule_interaction_parameters' : MetaInfoMap(startpage),
        'number_of_atoms_per_molecule_interaction' : MetaInfoMap(startpage),
        'number_of_molecule_interactions' : MetaInfoMap(startpage)
        }

    # section_atom_in_molecule of section_molecule_type
    atom_in_mol = {
        'atom_in_molecule_charge' : MetaInfoMap(startpage),
        'atom_in_molecule_name' : MetaInfoMap(startpage),
        'atom_in_molecule_to_atom_type_ref' : MetaInfoMap(startpage)
        }


    if defname == 'system':
        dictionary = systemDict
    elif defname == 'topology':
        dictionary = topology
    elif defname == 'singleconfcalc':
        dictionary = singleconfcalc
    elif defname == 'restrictions':
        dictionary = restrictions
    elif defname == 'frameseq':
        dictionary = frameseq
    elif defname == 'frameseqend':
        dictionary = frameseqend
    elif defname == 'atom_type':
        dictionary = atom_type
    elif defname == 'molecule_type':
        dictionary = moltypeDict
    elif defname == 'interaction':
        dictionary = interaction
    elif defname == 'sampling':
        dictionary = sampling
    elif defname == 'singlevdw':
        dictionary = singlevdw
    elif defname == 'confcore':
        dictionary = configuration_core
    else:
        dictionary = singleconfcalclist
    return MapDictionary(dictionary)





# Copyright 2015-2018 Franz Knuth, Fawzi Mohamed, Wael Chibani, Ankit Kariryaa, Lauri Himanen, Danio Brambila
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
try:
    import setup_paths
except ImportError:
    pass
import numpy as np
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.unit_conversion.unit_conversion import convert_unit
import json, os, re

############################################################
# This file contains functions that are needed
# by more than one parser.
############################################################

def get_metaInfo(filePath):
    """Loads metadata.

    Args:
        filePath: Location of metadata.

    Returns:
        metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    """
    metaInfoEnv, warnings = loadJsonFile(filePath = filePath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None)
    return metaInfoEnv

def write_controlIn(backend, metaInfoEnv, valuesDict, writeXC, location, logger):
    """Writes the values of the control.in file.

    Write the last occurrence of a keyword, i.e. [-1], since aims uses the last occurrence of a keyword.

    ATTENTION
    backend.superBackend is used here instead of only the backend to write the JSON values,
    since this allows to bybass the caching setting which was used to collect the values for processing.
    However, this also bypasses the checking of validity of the metadata name by the backend.
    The scala part will check the validity nevertheless.

    Args:
        backend: Class that takes care of writing and caching of metadata.
        metaInfoEnv: Loaded metadata.
        valuesDict: Dictionary that contains the cached values of a section.
        writeXC: Boolean that determines whether the keywords related to the xc functional should be written.
        location: A string that is used to specify where more than one setting for xc was found.
        logger: Logging object where messages should be written to.
    """
    # list of excluded metadata for writeout
    # k_grid is written with k1 so that k2 and k2 are not needed.
    # fhi_aims_controlIn_relativistic_threshold is written with fhi_aims_controlIn_relativistic.
    # The xc setting have to be handeled separatly since having more than one gives undefined behavior.
    # hse_omega is only written if HSE was used and converted according to hse_unit which is not written since not needed.
    # hybrid_xc_coeff is only written for hybrid functionals.
    # verbatim_writeout is only needed to detect if the control.in is written in the main file of aims.
    exclude_list = [
                    'x_fhi_aims_controlIn_k2',
                    'x_fhi_aims_controlIn_k3',
                    'x_fhi_aims_controlIn_relativistic_threshold',
                    'x_fhi_aims_controlIn_xc',
                    'x_fhi_aims_controlIn_hse_omega',
                    'x_fhi_aims_controlIn_hse_unit',
                    'x_fhi_aims_controlIn_hybrid_xc_coeff',
                    'x_fhi_aims_controlIn_verbatim_writeout',
                   ]
    # write settings
    for k,v in valuesDict.items():
        if k.startswith('x_fhi_aims_controlIn_'):
            if k in exclude_list:
                continue
            # write k_krid
            elif k == 'x_fhi_aims_controlIn_k1':
                write_k_grid(backend, 'x_fhi_aims_controlIn_k', valuesDict)
            elif k == 'x_fhi_aims_controlIn_relativistic':
                # check for scalar ZORA setting and convert to one common name
                if re.match(r"\s*zora\s+scalar", v[-1], re.IGNORECASE):
                    backend.superBackend.addValue(k, 'zora scalar')
                    # write threshold only for scalar ZORA
                    value = valuesDict[k + '_threshold']
                    if value is not None:
                        backend.superBackend.addValue(k + '_threshold', value[-1])
                else:
                    backend.superBackend.addValue(k, v[-1].lower())
            # default writeout
            else:
                # convert keyword values of control.in which are strings to lowercase for consistency
                if isinstance(v[-1], str):
                    value = v[-1].lower()
                else:
                    value = v[-1]
                backend.superBackend.addValue(k, value)
    # handling of xc functional
    if writeXC:
        write_xc_functional(backend = backend,
            metaInfoEnv = metaInfoEnv,
            metaNameStart = 'x_fhi_aims_controlIn',
            valuesDict = valuesDict,
            location = 'control.in',
            logger = logger)

def write_k_grid(backend, metaName, valuesDict):
    """Function to write k-grid for controlIn and controlInOut.

    Args:
        backend: Class that takes care of writing and caching of metadata.
        metaName: Corresponding metadata name for k.
        valuesDict: Dictionary that contains the cached values of a section.
    """
    k_grid = []
    for i in ['1', '2', '3']:
        ki = valuesDict.get(metaName + i)
        if ki is not None:
            k_grid.append(ki[-1])
    if k_grid:
        backend.superBackend.addArrayValues(metaName + '_grid', np.asarray(k_grid))

def write_xc_functional(backend, metaInfoEnv, metaNameStart, valuesDict, location, logger):
    """Function to write xc settings for controlIn and controlInOut.

    The omega of the HSE-functional is converted to the unit given in the metadata.
    The xc functional from controlInOut is converted and writen in the format specified in the metadata.

    Args:
        backend: Class that takes care of writing and caching of metadata.
        metaInfoEnv: Loaded metadata.
        metaNameStart: Base name of metdata for xc. Must be fhi_aims_controlIn or fhi_aims_controlInOut.
        valuesDict: Dictionary that contains the cached values of a section.
        location: A string that is used to specify where more than one setting for xc was found.
        logger: Logging object where messages should be written to.
    """
    # two functions to convert hybrid_xc_coeff to the correct weight
    def GGA_weight(x):
        return 1.0 - x
    def HF_weight(x):
        return x
    # TODO vdW functionals, double-hybrid functionals, screx, and cohsex
    # Dictionary for conversion of xc functional name in aims to metadata format.
    # The individual x and c components of the functional are given as dictionaries.
    # Possible keys of such a dictionary are 'name', 'weight', and 'convert'.
    # If 'weight' is not given it is not written.
    # With 'convert', a funtion is specified how hybrid_xc_coeff is converted to the correct weight for this xc component.
    xcDict = {
              'Perdew-Wang parametrisation of Ceperley-Alder LDA':   [{'name': 'LDA_C_PW'}, {'name': 'LDA_X'}],
              'Perdew-Zunger parametrisation of Ceperley-Alder LDA': [{'name': 'LDA_C_PZ'}, {'name': 'LDA_X'}],
              'VWN-LDA parametrisation of VWN5 form':                [{'name': 'LDA_C_VWN'}, {'name': 'LDA_X'}],
              'VWN-LDA parametrisation of VWN-RPA form':             [{'name': 'LDA_C_VWN_RPA'}, {'name': 'LDA_X'}],
              'AM05 gradient-corrected functionals':                 [{'name': 'GGA_C_AM05'}, {'name': 'GGA_X_AM05'}],
              'BLYP functional':                                     [{'name': 'GGA_C_LYP'}, {'name': 'GGA_X_B88'}],
              'PBE gradient-corrected functionals':                  [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_PBE'}],
              'PBEint gradient-corrected functional':                [{'name': 'GGA_C_PBEINT'}, {'name': 'GGA_X_PBEINT'}],
              'PBEsol gradient-corrected functionals':               [{'name': 'GGA_C_PBE_SOL'}, {'name': 'GGA_X_PBE_SOL'}],
              'RPBE gradient-corrected functionals':                 [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_RPBE'}],
              'revPBE gradient-corrected functionals':               [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_PBE_R'}],
              'PW91 gradient-corrected functionals':                 [{'name': 'GGA_C_PW91'}, {'name': 'GGA_X_PW91'}],
              'M06-L gradient-corrected functionals':                [{'name': 'MGGA_C_M06_L'}, {'name': 'MGGA_X_M06_L'}],
              'M11-L gradient-corrected functionals':                [{'name': 'MGGA_C_M11_L'}, {'name': 'MGGA_X_M11_L'}],
              'TPSS gradient-corrected functionals':                 [{'name': 'MGGA_C_TPSS'}, {'name': 'MGGA_X_TPSS'}],
              'TPSSloc gradient-corrected functionals':              [{'name': 'MGGA_C_TPSSLOC'}, {'name': 'MGGA_X_TPSS'}],
              'hybrid B3LYP functional':                             [{'name': 'HYB_GGA_XC_B3LYP5'}],
              'Hartree-Fock':                                        [{'name': 'HF_X'}],
              'HSE':                                                 [{'name': 'HYB_GGA_XC_HSE03'}],
              'HSE-functional':                                      [{'name': 'HYB_GGA_XC_HSE06'}],
              'hybrid-PBE0 functionals':                             [{'name': 'GGA_C_PBE'}, {'name': 'GGA_X_PBE', 'weight': 0.75, 'convert': GGA_weight}, {'name': 'HF_X', 'weight': 0.25, 'convert': HF_weight}],
              'hybrid-PBEsol0 functionals':                          [{'name': 'GGA_C_PBE_SOL'}, {'name': 'GGA_X_PBE_SOL', 'weight': 0.75, 'convert': GGA_weight}, {'name': 'HF_X', 'weight': 0.25, 'convert': HF_weight}],
              'Hybrid M06 gradient-corrected functionals':           [{'name': 'HYB_MGGA_XC_M06'}],
              'Hybrid M06-2X gradient-corrected functionals':        [{'name': 'HYB_MGGA_XC_M06_2X'}],
              'Hybrid M06-HF gradient-corrected functionals':        [{'name': 'HYB_MGGA_XC_M06_HF'}],
              'Hybrid M08-HX gradient-corrected functionals':        [{'name': 'HYB_MGGA_XC_M08_HX'}],
              'Hybrid M08-SO gradient-corrected functionals':        [{'name': 'HYB_MGGA_XC_M08_SO'}],
              'Hybrid M11 gradient-corrected functionals':           [{'name': 'HYB_MGGA_XC_M11'}],
             }
    # description for hybrid coefficient
    xcHybridCoeffDescr = 'hybrid coefficient $\\alpha$'
    # descritpion for omega of HSE06
    xcOmegaDescr = '$\\omega$ in m^-1'
    # distinguish between control.in and the aims output from the parsed control.in
    if metaNameStart == 'x_fhi_aims_controlIn':
        hseFunc = 'hse06'
        # functionals where hybrid_xc_coeff is written
        writeHybridCoeff = ['b3lyp', 'hse03', 'hse06', 'pbe0', 'pbesol0']
        xcWrite = False
    elif metaNameStart == 'x_fhi_aims_controlInOut':
        hseFunc = 'HSE-functional'
        # functionals where hybrid_xc_coeff is written
        writeHybridCoeff = ['hybrid B3LYP functional', 'HSE', 'HSE-functional', 'hybrid-PBE0 functionals', 'hybrid-PBEsol0 functionals']
        xcWrite = True
    else:
        logger.error("Unknown metaNameStart %s in function write_xc_functional in %s. Please correct." % (metaNameStart, os.path.basename(__file__)))
        return
    # get cached values for xc functional
    xc = valuesDict.get(metaNameStart + '_xc')
    if xc is not None:
        # check if only one xc keyword was found in control.in
        if len(xc) > 1:
            logger.error("Found %d settings for the xc functional in %s: %s. This leads to an undefined behavior of the calculation and no metadata can be written for %s_xc." % (len(xc), location, xc, metaNameStart))
        else:
            backend.superBackend.addValue(metaNameStart + '_xc', xc[-1])
            # check for hybrid_xc_coeff
            hybridCoeff = valuesDict.get(metaNameStart + '_hybrid_xc_coeff')
            # write hybrid_xc_coeff for certain functionals
            if hybridCoeff is not None and xc[-1] in writeHybridCoeff:
                backend.superBackend.addValue(metaNameStart + '_hybrid_xc_coeff', hybridCoeff[-1])
            # hse_omega is only written for HSE06
            if xc[-1] == hseFunc:
                hse_omega = valuesDict.get(metaNameStart + '_hse_omega')
                if hse_omega is not None:
                    # get unit from metadata
                    unit = metaInfoEnv.infoKinds[metaNameStart + '_hse_omega'].units
                    # try unit conversion and write hse_omega
                    hse_unit = valuesDict.get(metaNameStart + '_hse_unit')
                    if hse_unit is not None:
                        omegaValue = None
                        if hse_unit[-1] in ['a', 'A']:
                            omegaValue = convert_unit(hse_omega[-1], 'angstrom**-1', unit)
                        elif hse_unit[-1] in ['b', 'B']:
                            omegaValue = convert_unit(hse_omega[-1], 'bohr**-1', unit)
                        if omegaValue is not None:
                            backend.superBackend.addValue(metaNameStart + '_hse_omega', omegaValue)
                        else:
                            logger.warning("Unknown hse_unit %s in %s. Cannot write %s" % (hse_unit[-1], location, metaNameStart + '_hse_omega'))
                    else:
                        logger.warning("No value found for %s. Cannot write %s" % (metaNameStart + '_hse_unit', metaNameStart + '_hse_omega'))
            # convert xc functional for metadata
            if xcWrite:
                # get list of xc components according to parsed value
                xcList = xcDict.get(xc[-1])
                if xcList is not None:
                    # loop over the xc components
                    for xcItem in xcList:
                        xcName = xcItem.get('name')
                        if xcName is not None:
                            # write section and and XC_functional_name
                            gIndexTmp = backend.openSection('section_XC_functionals')
                            backend.addValue('XC_functional_name', xcName)
                            # write hybrid_xc_coeff for B3LYP and HSE03 into XC_functional_parameters
                            if hybridCoeff is not None and xc[-1] in ['hybrid B3LYP functional', 'HSE']:
                                backend.addValue('XC_functional_parameters', {xcHybridCoeffDescr: hybridCoeff[-1]})
                            # write omega and hybrid_xc_coeff for HSE06
                            elif xc[-1] == hseFunc:
                                # converted value of omega was obtained above
                                parameters = {xcOmegaDescr: omegaValue}
                                # add hybrid_xc_coeff
                                if hybridCoeff is not None:
                                    hybrid = hybridCoeff[-1]
                                else:
                                    hybrid = 0.25
                                parameters[xcHybridCoeffDescr] = hybrid
                                backend.addValue('XC_functional_parameters', parameters)
                            # adjust weight of functionals that are affected by hybrid_xc_coeff
                            elif hybridCoeff is not None and 'convert' in xcItem:
                                backend.addValue('XC_functional_weight', xcItem['convert'](hybridCoeff[-1]))
                            # write weight if present for current xcItem
                            else:
                                xcWeight = xcItem.get('weight')
                                if xcWeight is not None:
                                    backend.addValue('XC_functional_weight', xcWeight)
                            backend.closeSection('section_XC_functionals', gIndexTmp)
                        else:
                            logger.error("The dictionary for xc functional '%s' does not have the key 'name'. Please correct the dictionary xcDict in %s." % (xc[-1], os.path.basename(__file__)))
                else:
                    logger.error("The xc functional '%s' could not be converted for the metadata. Please add it to the dictionary xcDict in %s." % (xc[-1], os.path.basename(__file__)))
